"""
Metrics manager module to process request data
with prometheus client
"""

import glob
import os
import tempfile

if not (MULTIPROC_DIR := os.environ.get("PROMETHEUS_MULTIPROC_DIR")):
    TEMPDIR = tempfile.TemporaryDirectory()  # pylint: disable=R1732
    MULTIPROC_DIR = TEMPDIR.name
    os.environ["PROMETHEUS_MULTIPROC_DIR"] = TEMPDIR.name
if not os.environ.get("PROMETHEUS_DISABLE_CREATED_SERIES"):
    os.environ["PROMETHEUS_DISABLE_CREATED_SERIES"] = "true"

# pylint: disable=C0413
from prometheus_client import CollectorRegistry
from prometheus_client.multiprocess import MultiProcessCollector

from ..utils import mixins
from .entities import BaseMetricsContainer, MetricsSettings


class LunaMetricsManager(mixins.Initializable):
    """Metrics prometheus"""

    def __init__(self, settings: MetricsSettings, metricsContainer: BaseMetricsContainer):
        self.settings = settings
        self.registry = None
        self.container = metricsContainer

    async def initialize(self) -> None:
        """Initialize metrics collectors and storage"""
        if not self.settings.enabled:
            return
        self.registry = CollectorRegistry()
        if self.container.multiprocEnabled:
            self.prepareStorageDirectory()
            MultiProcessCollector(self.registry)

        self.container.initialize(self.registry)

    @staticmethod
    def prepareStorageDirectory() -> None:
        """Prepares directory to store metrics"""
        if not os.path.exists(MULTIPROC_DIR):
            os.mkdir(MULTIPROC_DIR)
        else:
            for file in glob.glob(MULTIPROC_DIR + "/*.db"):
                os.remove(file)

    async def close(self) -> None:
        """Storage cleanup"""
        if self.container.multiprocEnabled:
            for file in glob.glob(MULTIPROC_DIR + "/*.db"):
                os.remove(file)

    def handleRequest(self, request, response) -> None:
        """
        Parse request to prometheus metrics
        Args:
            request: request object
            response: response object
        """
        if not self.settings.enabled:
            return
        self.container.handleRequest(request, response)
