"""Module containing Clickhouse queries"""

from collections import namedtuple

BASIC_SERIES = ("requests", "errors")
NON_ESTIMATED_SERIES = (
    "requests",
    "errors",
    "sdkloop_task",
    "licensing",
    "ws_requests",
    "redis_message",
    "subscription_error",
    "matching-process",
    "subtask_final",
)

# estimators to use in the usages_statistics table
ESTIMATORS_TO_TRACK = (
    "face_detector_usages",
    "landmarks68_detector_usages",
    "head_pose_estimator_usages",
    "liveness_estimator_usages",
    "deepfake_estimator_usages",
    "mask_estimator_usages",
    "emotion_estimator_usages",
    "mouth_estimator_usages",
    "eye_estimator_usages",
    "gaze_estimator_usages",
    "glasses_estimator_usages",
    "face_warp_quality_estimator_usages",
    "face_basic_attributes_extractor_usages",
    "face_descriptor_extractor_usages",
    "body_detector_usages",
    "body_basic_attributes_estimator_usages",
    "body_upper_attributes_estimator_usages",
    "body_lower_attributes_estimator_usages",
    "body_accessories_estimator_usages",
    "body_descriptor_extractor_usages",
    "iso_estimator_usages",
    "face_quality_estimator_usages",
    "people_counter_usages",
)

BIN_SIZE = 50.0
MAX_MEANINGFUL_IMAGE_SIZE = 10000.0

INSERT_QUERY = "INSERT INTO {table} FORMAT JSONEachRow"
CREATE_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS {table} 
    (time DateTime64(6) NOT NULL, data Object(Nullable('json')) NOT NULL) 
    engine MergeTree ORDER BY time
"""

REQUESTS_COUNT_AGGREGATED_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS requests_count_daily
    (
        date Date NOT NULL, 
        service String NOT NULL, 
        status_code Int16 NOT NULL, 
        route String NOT NULL,
        count Int64
    )
    ENGINE SummingMergeTree
    PRIMARY KEY (date, service, route, status_code)
    ORDER BY (date, service, route, status_code);
"""

RESPONSE_TIME_AGGREGATED_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS {tableName}_aggregated_daily
    (
        date Date NOT NULL, 
        service String NOT NULL, 
        status_code Int16 NOT NULL, 
        route String NOT NULL,
        avg_time Float64 NOT NULL,
    )
    ENGINE AggregatingMergeTree
    PRIMARY KEY (date, service, route, status_code)
    ORDER BY (date, service, route, status_code);
"""

ERRORS_STATISTICS_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS errors_statistics_daily
    (
        date Date NOT NULL,
        service String NOT NULL, 
        error_code Int16 NOT NULL, 
        route String NOT NULL,
        count Int64
    )
    ENGINE SummingMergeTree
    PRIMARY KEY (date, service, route, error_code) 
    ORDER BY (date, service, route, error_code);
"""

USAGE_STATISTICS_TABLE_QUERY = f"""
    CREATE TABLE IF NOT EXISTS usages_statistics_daily
    (
        date Date NOT NULL,
        service String NOT NULL,
        {',\n\t\t'.join([f"{estimator} AggregateFunction(sum, Int64) NOT NULL" for estimator in ESTIMATORS_TO_TRACK])}
    )
    ENGINE AggregatingMergeTree
    PRIMARY KEY (date, service)
    ORDER BY (date, service);
"""

SDK_ESTIMATOR_BATCH_SIZE_STATISTICS_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS sdk_estimator_batch_size_statistics
    (
        date Date NOT NULL,
        estimator String NOT NULL,
        service String NOT NULL,
        batch_size Float64,
    )
    ENGINE AggregatingMergeTree
    PRIMARY KEY (date, estimator, service)
    ORDER BY (date, estimator, service);
"""

SDK_ESTIMATOR_EXECUTION_TIME_STATISTICS_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS sdk_estimator_execution_time_statistics
    (
        date Date NOT NULL,
        estimator String NOT NULL,
        service String NOT NULL,
        execution_time Float64,
    )
    ENGINE AggregatingMergeTree
    PRIMARY KEY (date, estimator, service)
    ORDER BY (date, estimator, service);
"""
SDK_FACE_DETECTION_HEIGHT_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS sdk_face_detection_height_statistics
    (
        date Date NOT NULL,
        height Float32 NOT NULL,
        count Int64,
    )
    ENGINE SummingMergeTree
    PRIMARY KEY (date, height)
    ORDER BY (date, height);
"""
IMAGE_SIZE_AGGREGATION_TABLE_QUERY = """
    CREATE TABLE IF NOT EXISTS sdk_image_size_statistics_daily 
    (
        date Date NOT NULL, 
        width Float32 NOT NULL, 
        height Float32 NOT NULL, 
        count Int64
    ) 
    ENGINE SummingMergeTree
    PRIMARY KEY (date, width, height) 
    ORDER BY (date, width, height);
"""


REQUESTS_COUNT_MV_QUERY = """
    CREATE MATERIALIZED VIEW IF NOT EXISTS requests_count_daily_mv TO requests_count_daily
    AS SELECT 

        toDate(time) as date, 
        JSONExtractString(toJSONString(data), 'route') as route, 
        JSONExtractString(toJSONString(data), 'service') as service, 
        JSONExtractInt(toJSONString(data), 'status_code') as status_code, 
        count() as count

    FROM requests 
    GROUP BY date, service, route, status_code
    ORDER BY date, service, route, status_code;
"""

RESPONSE_TIME_MV_QUERY = """
    CREATE MATERIALIZED VIEW IF NOT EXISTS {toTable}_aggregated_daily_mv TO {toTable}_aggregated_daily
    AS SELECT 
        toDate(time) as date, 
        JSONExtractString(toJSONString(data), 'route') as route, 
        JSONExtractString(toJSONString(data), 'service') as service, 
        JSONExtractInt(toJSONString(data), 'status_code') as status_code, 
        avg(JSONExtractFloat(toJSONString(data), 'execution_time')) as avg_time
    FROM {fromTable} 
    WHERE JSONExtractFloat(toJSONString(data), 'execution_time') IS NOT NULL
    GROUP BY date, service, route, status_code
    ORDER BY date, service, route, status_code;
"""

ERRORS_STATISTICS_MV_QUERY = """
    CREATE MATERIALIZED VIEW IF NOT EXISTS errors_statistics_daily_mv TO errors_statistics_daily
    AS SELECT 
        toDate(time) as date, 
        JSONExtractString(toJSONString(data), 'route') as route, 
        JSONExtractString(toJSONString(data), 'service') as service, 
        JSONExtractInt(toJSONString(data), 'error_code') as error_code, 
        count() as count 
    FROM errors 
    GROUP BY date, service, route, error_code
    ORDER BY date, service, route, error_code;
"""

USAGE_STATISTICS_MV_QUERY = f"""
    CREATE MATERIALIZED VIEW IF NOT EXISTS usages_statistics_daily_mv TO usages_statistics_daily
    AS SELECT
        toDate(time) as date,
        JSONExtractString(toJSONString(data), 'service') as service,
        {
            ',\n\t\t'.join([
                f"sumState(JSONExtractInt(toJSONString(data), '{estimator}')) as {estimator}"
                for estimator in ESTIMATORS_TO_TRACK
            ])
        }
    FROM usages_statistic
    GROUP BY ALL
    ORDER BY date, service;
"""

SDK_ESTIMATOR_EXECUTION_TIME_MV_QUERY = """
    CREATE MATERIALIZED VIEW IF NOT EXISTS sdk_estimator_execution_time_statistics_{fromTable}_mv 
    TO sdk_estimator_execution_time_statistics
    AS SELECT
        toDate(time) as date,
        JSONExtractString(toJSONString(data), 'service') as service,
        avg(JSONExtractFloat(toJSONString(data), 'execution_time')) as execution_time,
        '{fromTable}' as estimator
    FROM {fromTable}
    WHERE JSONExtractFloat(toJSONString(data), 'execution_time') IS NOT NULL
    GROUP BY date, estimator, service
    ORDER BY date, estimator, service;
"""

SDK_ESTIMATOR_BATCH_SIZE_MV_QUERY = """
    CREATE MATERIALIZED VIEW IF NOT EXISTS sdk_estimator_batch_size_statistics_{fromTable}_mv 
    TO sdk_estimator_batch_size_statistics
    AS SELECT
        toDate(time) as date,
        '{fromTable}' as estimator,
        JSONExtractString(toJSONString(data), 'service') as service,
        avg(toFloat64(JSONExtractInt(toJSONString(data), 'batch_size'))) as batch_size
    FROM {fromTable}
    GROUP BY date, estimator, service
    ORDER BY date, estimator, service;
"""

IMAGE_SIZE_AGGREGATION_MV_QUERY = """
    CREATE MATERIALIZED VIEW IF NOT EXISTS  sdk_image_size_statistics_daily_mv TO sdk_image_size_statistics_daily 
    AS SELECT 
        toDate(time) as date, 
        intDiv(JSONExtractFloat(toJSONString(data), 'width'), {binSize}) * {binSize} as width,
        if(
            JSONExtractInt(toJSONString(data), 'height') < {maxHeight}, 
            JSONExtractInt(toJSONString(data), 'height') % {binSize} > 0 ? 
            toFloat32(intDiv(JSONExtractInt(toJSONString(data), 'height'), {binSize}) * {binSize} + {binSize}) : 
            JSONExtractFloat(toJSONString(data), 'height'), 
            inf
        ) as height,
        count() as count  
    FROM image_load 
    GROUP BY date, width, height 
    ORDER BY date, width, height;
"""

SDK_FACE_DETECTION_HEIGHT_MV_QUERY = """
    CREATE MATERIALIZED VIEW IF NOT EXISTS sdk_face_detection_height_stats_{fromTable}_mv 
    TO sdk_face_detection_height_statistics
    AS SELECT 
        toDate(time) as date,
        if (
            JSONExtractFloat(toJSONString(data), 'height') < {maxHeight}, 
            JSONExtractFloat(toJSONString(data), 'height') % {binSize} > 0 ?
            toFloat32(intDiv(JSONExtractFloat(toJSONString(data), 'height'), {binSize}) * {binSize} + {binSize}) : 
            JSONExtractFloat(toJSONString(data), 'height'), 
            Inf
        ) as height, 
        count() as count  
    FROM {fromTable}
    GROUP BY date, height
    ORDER BY date, height;
"""

Aggregation = namedtuple("Aggregation", "createTable, createMaterializedView")

aggregations = {
    "requests": [
        Aggregation(
            REQUESTS_COUNT_AGGREGATED_TABLE_QUERY,
            REQUESTS_COUNT_MV_QUERY,
        ),
        Aggregation(
            RESPONSE_TIME_AGGREGATED_TABLE_QUERY.format(tableName="response_time"),
            RESPONSE_TIME_MV_QUERY.format(
                fromTable="requests",
                toTable="response_time",
            ),
        ),
        Aggregation(
            RESPONSE_TIME_AGGREGATED_TABLE_QUERY.format(tableName="response_time_p95"),
            RESPONSE_TIME_MV_QUERY.format(
                fromTable="requests",
                toTable="response_time_p95",
            ),
        ),
        Aggregation(
            RESPONSE_TIME_AGGREGATED_TABLE_QUERY.format(tableName="response_time_p99"),
            RESPONSE_TIME_MV_QUERY.format(
                fromTable="requests",
                toTable="response_time_p99",
            ),
        ),
    ],
    "errors": [Aggregation(ERRORS_STATISTICS_TABLE_QUERY, ERRORS_STATISTICS_MV_QUERY)],
    "usages_statistic": [Aggregation(USAGE_STATISTICS_TABLE_QUERY, USAGE_STATISTICS_MV_QUERY)],
    "image_load": [
        Aggregation(
            RESPONSE_TIME_AGGREGATED_TABLE_QUERY.format(tableName="image_load_time"),
            RESPONSE_TIME_MV_QUERY.format(
                fromTable="image_load",
                toTable="image_load_time",
            ),
        ),
        Aggregation(
            IMAGE_SIZE_AGGREGATION_TABLE_QUERY,
            IMAGE_SIZE_AGGREGATION_MV_QUERY.format(binSize=BIN_SIZE, maxHeight=MAX_MEANINGFUL_IMAGE_SIZE),
        ),
    ],
    "face_detection_height": Aggregation(
        SDK_FACE_DETECTION_HEIGHT_TABLE_QUERY,
        SDK_FACE_DETECTION_HEIGHT_MV_QUERY,
    ),
    "execution_time": Aggregation(
        SDK_ESTIMATOR_EXECUTION_TIME_STATISTICS_TABLE_QUERY, SDK_ESTIMATOR_EXECUTION_TIME_MV_QUERY
    ),
    "batch_size": Aggregation(SDK_ESTIMATOR_BATCH_SIZE_STATISTICS_TABLE_QUERY, SDK_ESTIMATOR_BATCH_SIZE_MV_QUERY),
}
