"""Fast Json and msg pack loader and validator."""

from typing import Callable

import fastjsonschema
from jsonschema_rs import ValidationError as RSValidationError, jsonschema_rs  # pylint: disable-msg=no-name-in-module
from vlutils.helpers import bytesToBase64

from ..errors.errors import Error
from ..errors.exception import VLException
from ..web.request import VLRequest
from ..web.utils import decodeBytes, isValidJsonRequestContentType, loadJson, loadMsgPack, validateJson


def _encodeObjectBytes(obj: dict | list) -> dict | list:
    """Msgpack hook for convering bytes to string"""
    if isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, bytes):
                obj[key] = bytesToBase64(value)
    if isinstance(obj, list):
        for idx, value in enumerate(obj):
            if isinstance(value, bytes):
                obj[idx] = bytesToBase64(value)
    return obj


def _loadData(request: VLRequest, convertBytes: bool = False) -> tuple[str, dict]:
    """
    Loads request data from json / msgpack.
    Args:
        convertBytes: whether to load data from msgpack with a custom hook to convert bytes
                      making data suitable for jsonschema-rs validation
    """

    if request.parsed_json is None:
        try:
            if request.content_type == "application/msgpack":
                request.parsed_json = loadMsgPack(
                    request.body,
                    object_hook=_encodeObjectBytes if convertBytes else None,
                    list_hook=_encodeObjectBytes if convertBytes else None,
                )
            elif isValidJsonRequestContentType(request.content_type):
                text = decodeBytes(body=request.body, encoding=request.charset)
                request.parsed_json = loadJson(text)
            else:
                raise VLException(Error.BadContentType, 400, isCriticalError=False)
            return request.content_type, request.parsed_json
        except ValueError as exc:
            raise VLException(Error.RequestNotContainsJson, 400, isCriticalError=False) from exc
    return request.content_type, request.parsed_json


class RSLoader:
    """
    Json and msg pack loader and validator.

    A validation is based on a rust jsonschema validator. This validator much faster but has some restrictions.

    Warnings:

        Known restrictions:
          * rust jsonschema validator does not support bytes in the schema. Loader can to convert bytes to base64 string
            and validate it such string (see params convertBytes). It is faster usually than validation with a
            fastjsonschema validator validation.
          * if a validation with the rust validator has failed we will re check loaded object with a fastjsonschema
            validator. It was made for error correct parsing
    Attributes:
        cpRSJsSchemaCache: compiled rust schema cache
        cpFJsSchemaCache: compiled fastjsonschema schema cache
        rsJsCompiler: rust schema compiler
        fJsCompiler: fastjsonschema schema compiler
    """

    def __init__(
        self,
        rstJsCompiler: Callable[[dict], Callable] | None = None,
        fJsCompiler: Callable[[dict], Callable] | None = None,
    ):
        self.cpRSJsSchemaCache = {}
        self.cpFJsSchemaCache = {}
        self.rsJsCompiler = rstJsCompiler if rstJsCompiler else jsonschema_rs.validator_for
        self.fJsCompiler = fJsCompiler if fJsCompiler else fastjsonschema.compile

    def load(
        self, request: VLRequest, jsSchema: dict, convertBytes: bool = False, revalidationSchema: dict | None = None
    ):
        """
        Load  json or msgpack to python dict and validate it

        Args:
            request: input request
            jsSchema: json schema for request
            convertBytes: convert bytes to base64 string or not (actual for msgpack
            revalidationSchema: revalidation schema for validation based on fastjsonschema
        Returns:
            loaded object
        """
        schemaId = id(jsSchema)
        cpJsSchema = self.cpRSJsSchemaCache.get(schemaId)
        if cpJsSchema is None:
            cpJsSchema = self.rsJsCompiler(jsSchema)
            self.cpRSJsSchemaCache[schemaId] = cpJsSchema
            fJsSchema = self.fJsCompiler(jsSchema)
            self.cpFJsSchemaCache[schemaId] = fJsSchema
        contentType, inputJson = _loadData(request, convertBytes)
        try:
            cpJsSchema.validate(inputJson)
            int("1")  # workaround for https://github.com/Stranger6667/jsonschema-rs/issues/435
        except (RSValidationError, SystemError, ValueError):
            # re-validation. May be false negative case + error generation through fastjsonschema validation
            # The underlying crate doesn’t support arbitrary precision integers yet, which may lead to SystemError
            # when such value is used;
            # https://pypi.org/project/jsonschema-rs/0.4.2/
            if revalidationSchema is None:
                validateJson(contentType, inputJson, self.cpFJsSchemaCache[schemaId], useJsonSchema=False)
            else:
                revalidationSchemaId = id(revalidationSchema)
                if (cpRevalidationSchema := self.cpFJsSchemaCache.get(revalidationSchemaId)) is None:
                    cpRevalidationSchema = self.fJsCompiler(revalidationSchema)
                    self.cpFJsSchemaCache[revalidationSchemaId] = cpRevalidationSchema
                validateJson(contentType, inputJson, cpRevalidationSchema, useJsonSchema=False)

        return inputJson
