"""
Module realizes service middlewares
"""

from gzip import compress as gcompress, decompress as gdecompress

try:
    from gzip import BadGzipFile
except ImportError:
    # fix for python 3.7
    BadGzipFile = OSError
from time import perf_counter
from typing import Optional, Union
from zlib import compress as zcompress, decompress as zdecompress, error as BadZlibFile

from sanic.response import HTTPResponse

from ..constants.http_consts import COMPRESS_MIN_SIZE
from ..errors.errors import Error, ErrorInfo
from ..errors.exception import VLException
from ..monitoring.points import DataForMonitoring, RequestErrorMonitoringPoint, RequestMonitoringPoint
from ..utils.log import logger
from ..web.request import VLRequest
from ..web.response import VLHTTPResponse


def logEndRequest(executionTime: float, statusCode: int):
    """
    Log of end a request.

    Args:
        executionTime: request execution time
        statusCode: status code
    """
    logger.info(
        f"End request, execution_time: {executionTime}, status code: {statusCode}",
        extra={"http.response.status_code": statusCode, "http.response.execution_time": executionTime},
    )


async def endRequestMiddleware(request: VLRequest, response: HTTPResponse):
    """
    End request middleware? print message about request ending, set output "Luna-Request-Id" header.
    Args:
        request: request
        response: response
    """
    executionTime = perf_counter() - request.startRequestTimeCheckPoint
    logEndRequest(executionTime, response.status)
    response.headers["Luna-Request-Id"] = request.requestId


def generateRequestMonitoringPoint(request: VLRequest, response: HTTPResponse) -> RequestMonitoringPoint:
    """
    Generate request monitoring point by request.

    Args:
        request: request
        response: response

    Returns:
        point
    """
    additionalMonitoringData: DataForMonitoring = request.dataForMonitoring

    point = RequestMonitoringPoint(
        requestTime=request.startRequestTime,
        statusCode=response.status,
        method=request.method,
        additionalFields=additionalMonitoringData.fields,
        additionalTags=additionalMonitoringData.tags,
        executionTime=perf_counter() - request.startRequestTimeCheckPoint,
        requestId=request.requestId,
        resource=request.beautyRoute,
        service=request.app.name,
    )
    return point


def getRequestErrorMonitoringPoint(
    request: VLRequest, response: HTTPResponse
) -> Union[RequestErrorMonitoringPoint, None]:
    """
    Generate request error monitoring point by request.

    Args:
        request: request
        response: response

    Returns:
        point if request has error
    """
    error = request.error
    if error is None:
        return None
    if not isinstance(error, ErrorInfo):
        return None
    additionalMonitoringData: DataForMonitoring = request.dataForMonitoring

    point = RequestErrorMonitoringPoint(
        requestTime=request.startRequestTime,
        statusCode=response.status,
        method=request.method,
        additionalFields=additionalMonitoringData.fields,
        additionalTags=additionalMonitoringData.tags,
        requestId=request.requestId,
        errorCode=error.errorCode,
        resource=request.beautyRoute,
        service=request.app.name,
    )
    return point


async def sendToMonitoringMiddleware(request: VLRequest, response: HTTPResponse):
    """
    Monitoring middleware. Generate a request point and error point and send them to app monitoring.

    Args:
        request: request
        response: response
    """

    requestPoint = generateRequestMonitoringPoint(request, response)
    requestErrorPoint = getRequestErrorMonitoringPoint(request, response)
    if requestErrorPoint is None:
        points = [requestPoint]
    else:
        points = [requestPoint, requestErrorPoint]

    request.appCtx.monitoring.flushPoints(points)
    if hasattr(request.appCtx.serviceConfig, "metrics"):
        request.appCtx.metrics.handleRequest(request, response)


def logResponseError(response: HTTPResponse, error: ErrorInfo):
    """
    log error from response.

    Args:
        response: response
        error: error which is occurred request processing
    """
    msg = (
        f"request failed, status code: {response.status}, reason: {error.errorCode}, desc: {error.description}, "
        f"detail: {error.detail}"
    )
    logger.info(
        msg,
        extra={
            "http.response.status_code": response.status,
            "error.code": error.errorCode,
        },
    )


async def logErrorMiddleware(  #: pylint: disable-msg=W0613
    request: Optional[VLRequest], response: Optional[HTTPResponse]
):
    """
    Middleware, that logs error response.

    Args:
        request: request
        response: response
    """
    if error := request.error:
        logResponseError(response, error)


async def compressResponseMiddleware(request: VLRequest, response: HTTPResponse):
    """
    Middleware that compress output response body according `Accept-Encoding` header

    Args:
        request: request
        response: response
    """
    if request.streamResponse:
        return
    uncompressedBodyLength = len(response.body)
    acceptEncoding = request.headers.get("Accept-Encoding")
    if not acceptEncoding:
        return
    if uncompressedBodyLength < COMPRESS_MIN_SIZE:
        response.headers["Content-Encoding"] = "identity"
        return
    contentType = response.headers.get("Content-Type")
    if contentType and (contentType.startswith("image/") or contentType.startswith("application/zip")):
        response.headers["Content-Encoding"] = "identity"
        return

    supposeContentEncoding = newContentEncoding = "identity"
    compressedBody = response.body
    if "deflate" in acceptEncoding:
        compressedBody = zcompress(response.body)
        supposeContentEncoding = "deflate"
    elif "gzip" in acceptEncoding:
        compressedBody = gcompress(response.body)
        supposeContentEncoding = "gzip"
    compressedBodyLength = len(compressedBody)
    if compressedBodyLength < uncompressedBodyLength or "identity" not in acceptEncoding:
        response.body = compressedBody
        response.headers["Vary"] = "Accept-Encoding"
        response.headers["Content-Length"] = compressedBodyLength
        newContentEncoding = supposeContentEncoding
    response.headers["Content-Encoding"] = newContentEncoding


async def deCompressResponseMiddleware(request: VLRequest):
    """
    Middleware that decompress input request body according `Content-Encoding` header

    Args:
        request: request
    """
    contentEncoding = request.headers.get("Content-Encoding")
    if not contentEncoding:
        return
    if not request.body:
        return
    if contentEncoding == "gzip":
        try:
            request.body = gdecompress(request.body)
        except BadGzipFile as exc:
            raise VLException(Error.InvalidInputGzip, 400, isCriticalError=False) from exc
    elif contentEncoding == "deflate":
        try:
            request.body = zdecompress(request.body)
        except BadZlibFile as exc:
            raise VLException(Error.InvalidInputDeflate, 400, isCriticalError=False) from exc


async def restMiddleware(request: VLRequest, response: VLHTTPResponse):
    """
    REST middleware. Middleware applies some REST rules to a request response.

    Args:
        request: request
        response: request response
    """
    statusCode = response.status
    if statusCode in (100, 101, 102, 103, 204, 304):
        # https://datatracker.ietf.org/doc/html/rfc7230#section-3.3
        if "Content-Encoding" in response.headers:
            del response.headers["Content-Encoding"]
        response.body = b""
    if statusCode in (100, 101, 102, 103, 204):
        if "Content-Length" in response.headers:
            del response.headers["Content-Length"]
    if request.method == "HEAD":
        # https://datatracker.ietf.org/doc/html/rfc7230#section-3.3
        response.headersFilter = lambda k, v: k.lower() != "content-length"
        response.body = b""
